/*--------------------------------------------------------------------------*/
// Sample Application													
//																			
// Chapter 14																
//																			
// Learn DirectX in 24 Hours												
// by Robert Dunlop															
//																			
// Copyright (C) 1999														
//
//
//  Chapter 14 code by Dave Barrett and Sam Christiansen
//  
/*--------------------------------------------------------------------------*/

//------ Include Files ------//

#define D3D_OVERLOADS
#define INITGUID
#include <windows.h>
#include <mmsystem.h>
#include <ddraw.h>
#include <d3d.h>
#include <d3dtypes.h>
#include "d3dutil.h"
#include "d3dmath.h"
#include "cube.h"

#define SafeRelease(x) if (x) { x->Release(); x=NULL;}
#define SafeDelete(x) if (x) {delete x;x=NULL;}

//------ Window Class Information ------//

static char szClass[] = "XmplHr14Class";
static char szCaption[] = "Example - Hour 14";

//------ Global Interface Pointers ------//

LPDIRECTDRAW7 lpDD=NULL;
LPDIRECTDRAWSURFACE7 lpDDSPrimary=NULL;
LPDIRECTDRAWSURFACE7 lpDDSBack=NULL;
LPDIRECT3D7 lpD3D=NULL;
LPDIRECT3DDEVICE7 lpDevice=NULL;

// Z-Buffer
LPDIRECTDRAWSURFACE7 lpDDSZBuf=NULL;

// Texture map structure
LPDIRECTDRAWSURFACE7 lpTexture=NULL;


// ---- Light height ------------------- //
float	lightElevation			= 200.0f;

//----- Rotation position and speed -----//

float rotAngle=g_PI;		// current angle
float rotVel=0.0f;			// current velociy of spin

//----- Elevation of viewer and vertical speed -----//

float elevation=350.0f;		// current elevation
float liftVel=0.0f;			// rate of rise / decent

//----- Define distance of viewer orbit from target ------//

#define ORBIT			800.0f

//------ Storage for Cube Objects -----//

#define NUM_ROWS		4
#define NUM_COLUMNS		4

CCube *cubes[NUM_ROWS][NUM_COLUMNS];

//------ Error Return String ------//

const char *ErrStr=NULL;	

//------ Error Messages ------//

const char Err_Reg_Class[]			= "Error Registering Window Class";
const char Err_Create_Win[]			= "Error Creating Window";
const char Err_DirectDrawCreate[]	= "DirectDrawCreate FAILED";
const char Err_Query[]				= "QueryInterface FAILED";
const char Err_Coop[]				= "SetCooperativeLevel FAILED";
const char Err_CreateSurf[]			= "CreateSurface FAILED";
const char Err_DispMode[]			= "Error Setting Display Mode";
const char Err_Device[]				= "Device Creation Failed";
const char Err_SetView[]			= "Viewport settings failed";

//------ Function Prototypes -----//

void Cleanup();
void create_objects();
static BOOL Init(HINSTANCE hInstance, int nCmdShow);
BOOL init_d3d();
BOOL init_ddraw(HWND hWnd);
void render_frame(float elapsed);
void create_lights();
void move_lights();
void toggle_light(int light_number);


// ----- Function Definitions -----//


//-----------------------------------------------------------------------------
// Name: TextureSearchCallback()
// Desc: Enumeration callback routine to find a 16-bit texture format. This
//       function is invoked by the ID3DDevice::EnumTextureFormats() function
//       to sort through all the available texture formats for a given device.
//       The pixel format of each enumerated texture format is passed into the
//       "pddpf" parameter. The 2nd parameter is to be used however the app
//       sees fit. In this case, we are using it as an output parameter to 
//       return a normal 16-bit texture format.
//-----------------------------------------------------------------------------
static HRESULT CALLBACK TextureSearchCallback( DDPIXELFORMAT* pddpf,
                                               VOID* param )
{
    // Note: Return with DDENUMRET_OK to continue enumerating more formats.

    // Skip any advanced modes
    if( pddpf->dwFlags & (DDPF_LUMINANCE|DDPF_BUMPLUMINANCE|DDPF_BUMPDUDV) )
        return DDENUMRET_OK;
    
    // Skip any FourCC formats
    if( pddpf->dwFourCC != 0 )
        return DDENUMRET_OK;

    // Skip alpha modes
    if( pddpf->dwFlags&DDPF_ALPHAPIXELS )
        return DDENUMRET_OK;

    // We only want 16-bit formats, so skip all others
    if( pddpf->dwRGBBitCount != 16 )
        return DDENUMRET_OK;

    // We found a good match. Copy the current pixel format to our output
    // parameter
    memcpy( (DDPIXELFORMAT*)param, pddpf, sizeof(DDPIXELFORMAT) );

    // Return with DDENUMRET_CANCEL to end enumeration.
    return DDENUMRET_CANCEL;
}

//-----------------------------------------------------------------------------
// Name: CreateTextureFromBitmap()
// Desc: Use a bitmap to create a texture for the specified device. This code
//       gets the attributes of the texture from the bitmap, creates the
//       texture, and then copies the bitmap into the texture.
//-----------------------------------------------------------------------------
static LPDIRECTDRAWSURFACE7 CreateTextureFromBitmap( LPDIRECT3DDEVICE7 pd3dDevice,
                                                     HBITMAP hbm )
{
    LPDIRECTDRAWSURFACE7 pddsTexture;
    HRESULT hr;

    // Get the device caps so we can check if the device has any constraints
    // when using textures (Voodoo cards for example, have a limit of 256x256 texture size)
    D3DDEVICEDESC7 ddDesc;
    if( FAILED( pd3dDevice->GetCaps( &ddDesc ) ) )
        return NULL;

    // Get the bitmap structure (to extract width, height, and bpp)
    BITMAP bm;
    GetObject( hbm, sizeof(BITMAP), &bm );
    DWORD dwWidth  = (DWORD)bm.bmWidth;
    DWORD dwHeight = (DWORD)bm.bmHeight;

    // Setup the new surface desc for the texture. Note how we are using the
    // texture manage attribute, so Direct3D does alot of dirty work for us
    DDSURFACEDESC2 ddsd;
    ZeroMemory( &ddsd, sizeof(DDSURFACEDESC2) );
    ddsd.dwSize          = sizeof(DDSURFACEDESC2);
    ddsd.dwFlags         = DDSD_CAPS|DDSD_HEIGHT|DDSD_WIDTH|
                           DDSD_PIXELFORMAT|DDSD_TEXTURESTAGE;
    ddsd.ddsCaps.dwCaps  = DDSCAPS_TEXTURE;
    ddsd.dwWidth         = dwWidth;
    ddsd.dwHeight        = dwHeight;

    // Turn on texture management for hardware devices
    if( ddDesc.deviceGUID == IID_IDirect3DHALDevice )
        ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
    else if( ddDesc.deviceGUID == IID_IDirect3DTnLHalDevice )
        ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
    else
        ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;

    // Adjust width and height, if the driver requires it
    if( ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_POW2 )
    {
        for( ddsd.dwWidth=1;  dwWidth>ddsd.dwWidth;   ddsd.dwWidth<<=1 );
        for( ddsd.dwHeight=1; dwHeight>ddsd.dwHeight; ddsd.dwHeight<<=1 );
    }
    if( ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY )
    {
        if( ddsd.dwWidth > ddsd.dwHeight ) ddsd.dwHeight = ddsd.dwWidth;
        else                               ddsd.dwWidth  = ddsd.dwHeight;
    }

    // Enumerate the texture formats, and find the closest device-supported
    // texture pixel format. The TextureSearchCallback function for this
    // tutorial is simply looking for a 16-bit texture. Real apps may be
    // interested in other formats, for alpha textures, bumpmaps, etc..
    pd3dDevice->EnumTextureFormats( TextureSearchCallback, &ddsd.ddpfPixelFormat );
    if( 0L == ddsd.ddpfPixelFormat.dwRGBBitCount )
        return NULL;

    // Get the device's render target, so we can then use the render target to
    // get a ptr to a DDraw object. We need the DirectDraw interface for
    // creating surfaces.
    LPDIRECTDRAWSURFACE7 pddsRender;
    LPDIRECTDRAW7        pDD;
    pd3dDevice->GetRenderTarget( &pddsRender );
    pddsRender->GetDDInterface( (VOID**)&pDD );
    pddsRender->Release();

    // Create a new surface for the texture
    if( FAILED( hr = pDD->CreateSurface( &ddsd, &pddsTexture, NULL ) ) )
    {
        pDD->Release();
        return NULL;
    }

    // Done with DDraw
    pDD->Release();

    // Now, copy the bitmap to the texture surface. To do this, we are creating
    // a DC for the bitmap and a DC for the surface, so we can use the BitBlt()
    // call to copy the actual bits.

    // Get a DC for the bitmap
    HDC hdcBitmap = CreateCompatibleDC( NULL );
    if( NULL == hdcBitmap )
    {
        pddsTexture->Release();
        return NULL;
    }
    SelectObject( hdcBitmap, hbm );

    // Get a DC for the surface
    HDC hdcTexture;
    if( SUCCEEDED( pddsTexture->GetDC( &hdcTexture ) ) )
    {
        // Copy the bitmap image to the surface.
        BitBlt( hdcTexture, 0, 0, bm.bmWidth, bm.bmHeight, hdcBitmap,
                0, 0, SRCCOPY );
        pddsTexture->ReleaseDC( hdcTexture );
    }
    DeleteDC( hdcBitmap );

    // Return the newly created texture
    return pddsTexture;
}




//-----------------------------------------------------------------------------
// Name: CreateTexture()
// Desc: Is passed a filename and creates a local Bitmap from that file. Some
//       logic and file parsing code could go here to support other image
//       file formats.
//-----------------------------------------------------------------------------
LPDIRECTDRAWSURFACE7 CreateTexture( LPDIRECT3DDEVICE7 pd3dDevice, 
                                    CHAR* strName )
{
    // Create a bitmap and load the texture file into it. Check the 
    // executable's resource first.
    HBITMAP hbm = (HBITMAP)LoadImage( GetModuleHandle(NULL), strName, 
                              IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );
    if( NULL == hbm )
    {
        // If not in the resource, try to load the bitmap as a file. Real code
        // would try to find the bitmap among many file paths.
        hbm = (HBITMAP)LoadImage( NULL, strName, IMAGE_BITMAP, 0, 0, 
                                  LR_LOADFROMFILE|LR_CREATEDIBSECTION );
        if( NULL == hbm )
            return NULL;
    }

    // The actual work of creating the texture is done in this next function.
    return CreateTextureFromBitmap( pd3dDevice, hbm );
}


void Cleanup()
{
	// de-allocate block objects

	for (int i=0;i<NUM_ROWS;i++)
		for (int j=0;j<NUM_COLUMNS;j++)
			SafeDelete(cubes[i][j]);

	// release 3D interfaces

	SafeRelease(lpDevice);
	SafeRelease(lpD3D);
	
	// release DirectDraw interfaces

	SafeRelease(lpDDSPrimary);
	SafeRelease(lpDD);

	// display error if one thrown

	if (ErrStr) {
		MessageBox(NULL, ErrStr, szCaption, MB_OK);
		ErrStr=NULL;
	}

}

LRESULT CALLBACK 
WindowProc(HWND hWnd, unsigned uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
		case WM_DESTROY:

			Cleanup();
            PostQuitMessage(0);
            break;

	    case WM_KEYDOWN: 

	        switch (wParam) { 

			case VK_UP:
			
				// move up

				liftVel=200.0f;
				break;

			case VK_DOWN:

				// move down

				liftVel=-200.0f;
				break;

			case VK_RIGHT:

				// rotate to the right

				rotVel=0.5f;
				break;

			case VK_LEFT:

				// rotate to the left

				rotVel=-0.5f;
				break; 
			}
			break;

	    case WM_KEYUP: 

	        switch (wParam) { 


			case '1':
				toggle_light(0);
				break;
			case '2':
				toggle_light(1);
				break;
			case '3':
				toggle_light(2);
				break;

			case 'l':case 'L':
				{
					DWORD on;
					lpDevice->GetRenderState(D3DRENDERSTATE_LIGHTING, &on);

					if (on)
						lpDevice->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE);
					else
						lpDevice->SetRenderState(D3DRENDERSTATE_LIGHTING, TRUE);
				}			
				break;


			case VK_UP:
			case VK_DOWN:
			
				liftVel=0.0f;
				break;

			case VK_RIGHT:
			case VK_LEFT:

				// rotate to the left

				rotVel=0.0f;
				break; 

			case VK_ESCAPE:

				// exit the program on escape

				DestroyWindow(hWnd);
				break;


			}
			break;

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
	
	return 0L;
}

BOOL init_ddraw(HWND hWnd)
{
	HRESULT ddrval;

    // Create the main DirectDraw object

    ddrval = DirectDrawCreateEx(NULL, (void **) &lpDD, IID_IDirectDraw7, NULL);
    if (ddrval != DD_OK) {
		ErrStr=Err_DirectDrawCreate;
		return FALSE;
	}

    // Set our cooperative level

    ddrval = lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
    if (ddrval != DD_OK) {
		ErrStr=Err_Coop;
		return FALSE;
	}

	// Set the display mode

	ddrval = lpDD->SetDisplayMode( 640, 480, 16, 0, 0);
	if (ddrval !=DD_OK) {
		ErrStr=Err_DispMode;
		return FALSE;
	}

    // Create the primary surface with 1 back buffer

    DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof( ddsd );
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                          DDSCAPS_FLIP | DDSCAPS_3DDEVICE |
                          DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 1;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
	if (ddrval!=DD_OK) {
		ErrStr=Err_CreateSurf;
		return FALSE;
	}

	// Fetch back buffer interface

	DDSCAPS2 ddscaps;
	ZeroMemory(&ddscaps,sizeof(ddscaps));
	ddscaps.dwCaps=DDSCAPS_BACKBUFFER;
	ddrval=lpDDSPrimary->GetAttachedSurface(&ddscaps,&lpDDSBack);
	if (ddrval!=DD_OK) {
		ErrStr=Err_CreateSurf;
		return FALSE;
	}

	// return success to caller

	return TRUE;
}

// Variables used when looking for z-buffer
BOOL zbuf_flag=FALSE;

// Used to keep track of which z-buffer device we got.
char zbuf_format=0;

// Z-Buffer callback function
static HRESULT WINAPI EnumZBufferFormatsCallback( DDPIXELFORMAT* pddpf,
                                                  VOID* pddpfDesired )
{
    // If parameters == NULL, don't enumerate more
	if( NULL==pddpf || NULL==pddpfDesired )
        return D3DENUMRET_CANCEL;

    // If the current pixel format's match the desired ones (DDPF_ZBUFFER and
    // possibly DDPF_STENCILBUFFER), lets copy it and return. This function is
    // not choosy...it accepts the first valid format that comes along.
    if( pddpf->dwFlags == ((DDPIXELFORMAT*)pddpfDesired)->dwFlags )
    {
        memcpy( pddpfDesired, pddpf, sizeof(DDPIXELFORMAT) );
		
		// Set flag to TRUE, since we got a valid z-buffer format.
		zbuf_flag=TRUE;
        return D3DENUMRET_CANCEL;
    }

    return D3DENUMRET_OK;
}

BOOL init_d3d()
{
	
	// Get Direct3D interface
	if (FAILED(lpDD->QueryInterface(IID_IDirect3D7, (LPVOID *)&lpD3D))) 
	{
		// Set error string.
		ErrStr=Err_Query;
		
		// Return false
		return FALSE;
	}
	
	// Enumerate the Z-buffer formats, making sure they fit the device.
	// First try hardware with accelerated transform and lighting.
	DDPIXELFORMAT m_ddpfZBuffer;
	ZeroMemory(&m_ddpfZBuffer,sizeof(m_ddpfZBuffer));
	m_ddpfZBuffer.dwSize=sizeof(m_ddpfZBuffer);
	m_ddpfZBuffer.dwFlags = DDPF_ZBUFFER ;
	lpD3D->EnumZBufferFormats( IID_IDirect3DTnLHalDevice, 
							   EnumZBufferFormatsCallback,
							   (VOID*)&m_ddpfZBuffer);
	
	
	if (zbuf_flag) {
		zbuf_format=1;
	} else {
		lpD3D->EnumZBufferFormats( IID_IDirect3DHALDevice, EnumZBufferFormatsCallback,
		                            (VOID*)&m_ddpfZBuffer );
		if (zbuf_flag) {
			zbuf_format=2;
		} else {
			lpD3D->EnumZBufferFormats( IID_IDirect3DMMXDevice, EnumZBufferFormatsCallback,
			                            (VOID*)&m_ddpfZBuffer );
			if (zbuf_flag) {
				zbuf_format=3;
			} else {
				lpD3D->EnumZBufferFormats( IID_IDirect3DRGBDevice, EnumZBufferFormatsCallback,
				                            (VOID*)&m_ddpfZBuffer );
				if (zbuf_flag) 
					zbuf_format=4;
			}
		}
	}

	// Create the zbuffer
	DDSURFACEDESC2 ddsd;
	HRESULT ddrval;
	ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof( ddsd );
    ddsd.dwFlags = DDSD_CAPS|DDSD_WIDTH|DDSD_HEIGHT|DDSD_PIXELFORMAT;
	
	// Use counter to check if we should create z-buffer in video or system memory
	// Note flag that specifies that this is a z-buffer surface.
	if (zbuf_format<3)
		ddsd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER|DDSCAPS_VIDEOMEMORY;
	else
		ddsd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER|DDSCAPS_SYSTEMMEMORY;
	
	// Set this to size of screen
	ddsd.dwWidth=640;
	ddsd.dwHeight=480;
	ddsd.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	ddsd.ddpfPixelFormat.dwFlags=DDPF_ZBUFFER;
	memcpy(&ddsd.ddpfPixelFormat,&m_ddpfZBuffer,sizeof(DDPIXELFORMAT));
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSZBuf, NULL );
	if (ddrval!=DD_OK) 
	{
		return FALSE; // Could not get z-buffer. Return false
	} else 
	{
		
		// Attach z-buffer to surface
		lpDDSBack->AddAttachedSurface(lpDDSZBuf);
	}

	// Set up the best device interface

	if (lpD3D->CreateDevice(IID_IDirect3DTnLHalDevice,lpDDSBack,&lpDevice)!=D3D_OK) 
	{
		if (lpD3D->CreateDevice(IID_IDirect3DHALDevice,lpDDSBack,&lpDevice)!=D3D_OK) 
		{
			if (lpD3D->CreateDevice(IID_IDirect3DMMXDevice,lpDDSBack,&lpDevice)!=D3D_OK) 
			{
				if (lpD3D->CreateDevice(IID_IDirect3DRGBDevice,lpDDSBack,&lpDevice)!=D3D_OK) 
				{
				// Set error string.
				ErrStr=Err_Device;
		
				// Return false
				return FALSE;
				}
			}
		}
	}

	// Set up the viewport
	D3DVIEWPORT7 view;
	view.dwX=0;
	view.dwY=0; 
	view.dwWidth=640;
	view.dwHeight=480; 
	view.dvMinZ=0.0f;
	view.dvMaxZ=1.0f;
	if (lpDevice->SetViewport(&view)!=D3D_OK)
	{
		// Set error string.
		ErrStr=Err_SetView;
		
		// Return false
		return FALSE;
	}
	
	// Projection matrix.
	D3DMATRIX proj_m;
	
	// Set projection matrix. 
	D3DUtil_SetProjectionMatrix(proj_m, 0.8f, 1.333f, 10.0f, 2000.0f );
	lpDevice->SetTransform(D3DTRANSFORMSTATE_PROJECTION,&proj_m);

	
	// Set the ambient light.
	D3DCOLOR d3dclrAmbientLightColor = D3DRGBA(0.1f, 0.1f, 0.1f, 0.1f);
	lpDevice->SetRenderState(D3DRENDERSTATE_AMBIENT, d3dclrAmbientLightColor);


	
	// Set linear (nice) filtering.
	lpDevice->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTFG_LINEAR);

	// Enable z-buffering. 
    lpDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, TRUE);


	// return success to caller
	
	return TRUE;
}

void create_objects()
{
	// Load texture
	lpTexture=CreateTexture(lpDevice,"building.bmp");
	
	// create cubes on a grid
	for (int i=0;i<NUM_ROWS;i++) 
		for (int j=0;j<NUM_COLUMNS;j++) 
		{
			cubes[i][j]=new CCube(D3DVECTOR(125.0f-100.0f*j,0.0f,125.0f-100.0f*i),
								  D3DVECTOR(50.0f,150.0f,50.0f),
								  0.95f,1.0f,0.8f);

			// Setup cubes material
			cubes[i][j]->SetMaterialAmbient(1.0f, 1.0f, 1.0f, 1.0f);
			cubes[i][j]->SetMaterialDiffuse(0.7f, 0.7f, 0.7f, 1.0f);
			cubes[i][j]->SetMaterialSpecular(1.0f, 1.0f, 1.0f, 1.0f, 80);

		}
}


void create_lights()
{ 
	//
	//	Note: Maximum number lights can be retrieved from the IDirect3DDevice7::GetCaps 
	//  function  ( dwMaxActiveLights member of LPD3DDEVICEDESC7  structure )
	//
	D3DLIGHT7 light0, light1, light2;

	// Enable Direct3D lighting.  
	lpDevice->SetRenderState(D3DRENDERSTATE_LIGHTING, TRUE); 
 	
	//
	//  Enable lights we are going to use 
	//
    lpDevice->LightEnable( 0, FALSE );
    lpDevice->LightEnable( 1, FALSE );
    lpDevice->LightEnable( 2, FALSE );
 
    ////////////////////////////////////////
	// Light 0 -- Initialize the structure.
    ZeroMemory(&light0, sizeof(D3DLIGHT7));

    
	// Position the light above the cubes in the scene
    //   Note: Lights use world space coordinates
    light0.dvPosition.x	 = 0.0f;
    light0.dvPosition.y  = lightElevation;
    light0.dvPosition.z  = 0.0f;
	
    // Set up for a mostly red point light.
    light0.dltType		 = D3DLIGHT_POINT;
    light0.dcvDiffuse.r  = 0.8f;
    light0.dcvDiffuse.g  = 0.2f;
    light0.dcvDiffuse.b  = 0.2f;
    light0.dcvAmbient.r  = 0.0f;
    light0.dcvAmbient.g  = 0.0f;
    light0.dcvAmbient.b  = 0.0f;
    light0.dcvSpecular.r = 1.0f;
    light0.dcvSpecular.g = 1.0f;
    light0.dcvSpecular.b = 1.0f;

	 // Don't attenuate.
    light0.dvAttenuation0 = 1.0f; 
    light0.dvRange		  = D3DLIGHT_RANGE_MAX;

	// Set the light in d3d
	lpDevice->SetLight(0, &light0);


	////////////////////////////////////////
	// Now on to light1 -- spotlight
    ZeroMemory(&light1, sizeof(D3DLIGHT7));

    light1.dvPosition.x  = 0.0f;
    light1.dvPosition.y  = lightElevation;
    light1.dvPosition.z  = 0.0f;

	// Set up for a mostly green spot light.
    light1.dltType		 = D3DLIGHT_SPOT;
    light1.dcvDiffuse.r  = 0.2f;
    light1.dcvDiffuse.g  = 0.8f;
    light1.dcvDiffuse.b  = 0.2f;
    light1.dcvAmbient.r  = 0.0f;
    light1.dcvAmbient.g  = 0.0f;
    light1.dcvAmbient.b  = 0.0f;
    light1.dcvSpecular.r = 1.0f;
    light1.dcvSpecular.g = 1.0f;
    light1.dcvSpecular.b = 1.0f;

	// Don't attenuate.
    light1.dvRange		  = D3DLIGHT_RANGE_MAX;


	light1.dvDirection    =  D3DVECTOR(  0,  -1,  0);			// point directly downward (in the -y direction)
    light1.dvFalloff      =  1.0f;								// linear falloff
    light1.dvTheta        =  3.14f/ 3.0f;						// inner ring ( 1/3 *PI radians)
    light1.dvPhi          =  2*3.14f /3.0f;						// outer ring ( 2/3 *PI radians)
    light1.dvAttenuation0 =  1.0f;								// Doesn't attenuate but still limited by range  (falloff)


	// Set the light in d3d
	lpDevice->SetLight(1, &light1 );

	////////////////////////////////////////
	// Now on to light2 -- directional
    ZeroMemory(&light2, sizeof(D3DLIGHT7));
	

	// Note Position is not used in directional lighting
	// just the Direction of the light 
    light2.dvDirection.x  = 0.0f;
    light2.dvDirection.y  = -1.0f;
    light2.dvDirection.z  = 0.0f;



    // Set up for a mostly blue directional light
    light2.dltType		 = D3DLIGHT_DIRECTIONAL;
    light2.dcvDiffuse.r  = 0.2f;
    light2.dcvDiffuse.g  = 0.2f;
    light2.dcvDiffuse.b  = 0.8f;
    light2.dcvAmbient.r  = 0.0f;
    light2.dcvAmbient.g  = 0.0f;
    light2.dcvAmbient.b  = 0.0f;
    light2.dcvSpecular.r = 1.0f;
    light2.dcvSpecular.g = 1.0f;
    light2.dcvSpecular.b = 1.0f;


	// Set the light in d3d
	lpDevice->SetLight(2, &light2 );
}

void move_lights()
{
	////////////////////////////////////////
	// constants for the function
	// 
	const  D3DVALUE radius		= 150.0f;			// radius of the circular path the light follows
	const  D3DVALUE rads_sec	= 3.14f /4.0f;		// change in angle in radians per second ( approx. PI/4 -> 1/8 turn )
	
	static D3DVALUE angle0		= 0;				// current angle in radians for light0  (base at 0 radians)
	static D3DVALUE angle1		= 2*3.14f /3.0f;	// current angle in radians for light1	(base 1/3 of the way 2/3*pi radians)
	static D3DVALUE angle2		= 4*3.14f /3.0f;	// current angle in radians for light2  (base 2/3 of the way 4/3*pi radians)

	static DWORD	last_time	= timeGetTime();	// last time we moved light

	//
	//	Calcluate angle change -	(# of millisecs / 1000 ) * radians per second = radians
	//
	DWORD		current_time	= timeGetTime();
	D3DVALUE	angle_delta		= ( (current_time - last_time) / 1000.0f )  *rads_sec;

	// Add angle for next frame for each of the light angles
	angle0	+= angle_delta;
	angle1	+= angle_delta;
	angle2	+= angle_delta;

	D3DLIGHT7 light;

	//
	// Calculate new position based on polar (x', y')  = ( r*cos(theta), r*sin(theta) )
	// Do it for each light - Equally space the lights around the circle
	//
	lpDevice->GetLight( 0, &light );
	light.dvPosition.x = radius * (D3DVALUE) cos( angle0 );
	light.dvPosition.z = radius * (D3DVALUE) sin( angle0 );
	light.dvPosition.y = lightElevation;
	lpDevice->SetLight( 0, &light );						// Set/Change the lights state
	
	lpDevice->GetLight( 1, &light );
	light.dvPosition.x = radius * (D3DVALUE) cos( angle1 );
	light.dvPosition.z = radius * (D3DVALUE) sin( angle1 );
	light.dvPosition.y = lightElevation;
	lpDevice->SetLight( 1, &light );						// Set/Change the lights state
	
	////////////////////////////////////////
	// The direction vector is moved not the position vector for directional lights
	lpDevice->GetLight( 2, &light );
	light.dvDirection.x = -1 * (D3DVALUE) cos( angle2 );	// Multiply by -1 to get the correct direction vector
	light.dvDirection.z = -1 * (D3DVALUE) sin( angle2 );	//	- All 3 lights will be equally spaced around the circle
	light.dvDirection.y = 0;								//Always aim through y=0 regardless of x and z 
	lpDevice->SetLight( 2, &light );						// Set/Change the lights state
	

	last_time	= current_time;								// store the time for the next pass
}


////////////////////////////////////////////////////////////////////////////////
// toggle_lights
//		-- Turns a specific light on or off. 
//
void toggle_light( int light_number )
{
	BOOL on;

	// Is the light enabled/on ?
	lpDevice->GetLightEnable( light_number, &on );

	// If it is, disable it, otherwise turn it on
	if (on)
		lpDevice->LightEnable( light_number, FALSE);
	else
		lpDevice->LightEnable( light_number, TRUE);
}



void render_frame(float elapsed)
{
	// recover any lost surfaces

	if (lpDDSPrimary->IsLost()==DDERR_SURFACELOST)
		lpDDSPrimary->Restore();
	if (lpDDSBack->IsLost()==DDERR_SURFACELOST)
		lpDDSBack->Restore();
	if (lpDDSZBuf->IsLost()==DDERR_SURFACELOST)
		lpDDSZBuf->Restore();
	if (lpTexture->IsLost()==DDERR_SURFACELOST)
		lpTexture->Restore();

	



	// increment viewer position
	elevation+=liftVel*elapsed;
	rotAngle+=rotVel*elapsed;

	// calculate current viewer position

	D3DVECTOR view_loc;
	view_loc.y=elevation;				// set elevation
	view_loc.x=sinf(rotAngle)*ORBIT;		// calculate position on x,z plane
	view_loc.z=cosf(rotAngle)*ORBIT;

	// create and set the view matrix

	D3DMATRIX view_matrix;
	D3DUtil_SetViewMatrix(view_matrix,
						  view_loc,
						  D3DVECTOR(0.0f,0.0f,0.0f),
						  D3DVECTOR(0.0f,1.0f,0.0f));
	lpDevice->SetTransform(D3DTRANSFORMSTATE_VIEW,&view_matrix);

 	// Clear the viewport. This time remembering to clear the z-buffer also
    lpDevice->Clear(0,NULL,D3DCLEAR_ZBUFFER|D3DCLEAR_TARGET,0,1.0f,0);

	// Set texture as active
	lpDevice->SetTexture( 0, lpTexture);
		
	// Start the scene render
	if( SUCCEEDED( lpDevice->BeginScene() ) ) 
	{
		
		move_lights();

		// Loop through the cubes and draw
		for (int i=0;i<NUM_ROWS;i++) 
			for (int j=0;j<NUM_COLUMNS;j++) 
				if (cubes[i][j])
					cubes[i][j]->draw(lpDevice);

		// end the scene

		lpDevice->EndScene();
	}

	// flip to the primary surface

	lpDDSPrimary->Flip(0,DDFLIP_WAIT);
}

//------ Function to Initialize DirectDraw and the Application ------//

static BOOL Init(HINSTANCE hInstance, int nCmdShow)
{
    WNDCLASS                    wc;

    // Set up and register window class

    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (WNDPROC) WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof(DWORD);
    wc.hInstance = hInstance;
    wc.hIcon = NULL;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = szClass;
    if (!RegisterClass(&wc)) {
		ErrStr=Err_Reg_Class;
        return FALSE;
	}

    // Get dimensions of display

    int ScreenWidth = GetSystemMetrics(SM_CXSCREEN);
    int ScreenHeight = GetSystemMetrics(SM_CYSCREEN);

    // Create a window and display

	HWND hWnd;

    hWnd = CreateWindow(szClass,							// class
                        szCaption,							// caption
						WS_VISIBLE|WS_POPUP,				// style 
						0,									// left
						0,									// top
						ScreenWidth,						// width
						ScreenHeight,						// height
                        NULL,								// parent window
                        NULL,								// menu 
                        hInstance,							// instance
                        NULL);								// parms
    if (!hWnd) {
		ErrStr=Err_Create_Win;
        return FALSE;
	}
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);

	// initialze DirectDraw

	if (!init_ddraw(hWnd)) return FALSE;

	// initialize Direct3D

	if (!init_d3d()) return FALSE;

	// create 3D objects

	create_objects();

	create_lights();


	// return success to caller

	return TRUE;
}

//------ Application Loop ------//

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	LONGLONG cur_time;       // current time
	LONGLONG perf_cnt;       // performance timer frequency
	BOOL perf_flag=FALSE;    // flag determining which timer to use
	LONGLONG last_time=0;	 // time of previous frame
	float time_elapsed;	 // time since previous frame
	float time_scale;		 // scaling factor for time

	// initialize the application, exit on failure

    if (!Init(hInstance, nCmdShow)) {
		Cleanup();
        return FALSE;
	}

	// is there a performance counter available? 

	if (QueryPerformanceFrequency((LARGE_INTEGER *) &perf_cnt)) { 

		// yes, set timer info and get starting time

		perf_flag=TRUE;
		QueryPerformanceCounter((LARGE_INTEGER *) &last_time); 
		time_scale=1.0f/perf_cnt;

	} else { 

		// no performance counter, read in using timeGetTime 

		last_time=timeGetTime(); 
		time_scale=0.001f;
	} 

    // Now we're ready to recieve and process Windows messages.

    BOOL bGotMsg;
    MSG  msg;
    PeekMessage( &msg, NULL, 0U, 0U, PM_NOREMOVE );
	
    while( WM_QUIT != msg.message  )
    {
        bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
        if( bGotMsg )
        {
			TranslateMessage( &msg );
            DispatchMessage( &msg );
        } else {
			
			// use the appropriate method to get time 
			// and calculate elapsed time since last frame

			if (perf_flag) 
				QueryPerformanceCounter((LARGE_INTEGER *) &cur_time); 
			else 
				cur_time=timeGetTime(); 

			// calculate elapsed time

			time_elapsed=(cur_time-last_time)*time_scale;

			// save frame time

			last_time=cur_time;
				
			// render the frame
			
			render_frame(time_elapsed);
		}
    }

	// return final message

    return msg.wParam;
}


